import pytest

from _testutils import MockHidapiDevice, MockRuntimeStorage
from liquidctl.driver.coolit import Coolit


class MockDevice(MockHidapiDevice):
    def read(self, length, **kwargs):
        return [0] * length


@pytest.fixture
def mock_h110i_gt():
    mock_raw_dev = MockDevice(vendor_id=0xDEAD, product_id=0xBEEF, address="42")
    mock_storage = MockRuntimeStorage(key_prefixes=["mock_h110i_gt"])
    dev = Coolit(
        mock_raw_dev,
        "Mock H100i GT",
        fan_count=2,
        rgb_fans=False,
    )
    return dev.connect(runtime_storage=mock_storage)


def test_driver_not_totally_broken(mock_h110i_gt):
    cooler = mock_h110i_gt
    cooler.initialize()
    _ = cooler.get_status(pump_mode="extreme")
    cooler.set_fixed_speed("fan1", 42)
    cooler.set_speed_profile("fan2", [(20, 30), (40, 90)])



"""liquidctl drivers for MSI liquid coolers.

Supported devices:

- MPG Coreliquid K360

Copyright (C) 2021  Andrew Udvare, Aapo Kössi and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

# uses the psf/black style

from collections import namedtuple
from collections.abc import Sequence
from copy import copy
from enum import Enum, unique
from time import sleep
import logging
import io
from PIL import Image

from liquidctl.driver.usb import UsbHidDriver
from liquidctl.keyval import RuntimeStorage
from liquidctl.util import RelaxedNamesEnum, check_unsafe, clamp, u16le_from

_LOGGER = logging.getLogger(__name__)

EXTRA_USAGE_PAGE = 0x0001
_MAX_DATA_LENGTH = 185
_PER_LED_LENGTH = 720
_REPORT_LENGTH = 64
_MAX_DUTIES = 7
_RAD_FAN_COUNT = 3
_CYCLE_NUMBER_STRIPE_TYPE_MAPPING = {0: 41, 1: 52, 2: 63, 3: 20, 4: 30}
# fmt: off
_DEFAULT_FEATURE_DATA = [
     82,   1, 255,   0,   0,  40,   0, 255,
      0, 128,   0,   1, 255,   0,   0,  40,
      0, 255,   0, 128,   0,   1, 255,   0,
      0,  40,   0, 255,   0, 128,   0,   1,
    255,   0,   0,  40,   0, 255,   0, 128,
      0,  20,   1, 255,   0,   0,  40,   0,
    255,   0, 128,   0,  20,   1, 255,   0,
      0,  40,   0, 255,   0, 130,  76,  10,
      1, 255,   0,   0,  40,   0, 255,   0,
    128,   0,  26, 255,   0,   0, 168,   0,
    255,   0, 191,   0,  32, 255,   0,   0,
     40,   0, 255,   0, 128,   0,  32, 255,
      0,   0,  40,   0, 255,   0, 128,   0,
     32, 255,   0,   0,  40,   0, 255,   0,
    128,   0,  32, 255,   0,   0,  40,   0,
    255,   0, 128,   0,  32, 255,   0,   0,
     40,   0, 255,   0, 128,   0,  32, 255,
      0,   0,  40,   0, 255,   0, 128,   0,
     32, 255,   0,   0,  40,   0, 255,   0,
    128,   0,  32, 255,   0,   0,  40,   0,
    255,   0, 128,   0,  32, 255,   0,   0,
     40,   0, 255,   0, 128,   0,  32, 255,
      0,   0,  40,   0, 255,   0, 128,   0,
      0
]
# fmt: on
_DeviceSettings = namedtuple(
    "DeviceSettings",
    [
        "stripe_or_fan",
        "fan_type",
        "corsair_device_quantity",
        "ll120_outer_individual",
        "led_num_jrainbow1",
        "led_num_jrainbow2",
        "led_num_jcorsair",
    ],
)
_BoardSyncSettings = namedtuple(
    "BoardSyncSettings",
    [
        "onboard_sync",
        "combine_jrgb",
        "combine_jpipe1",
        "combine_jpipe2",
        "combine_jrainbow1",
        "combine_jrainbow2",
        "combine_jcorsair",
    ],
)
_StyleSettings = namedtuple(
    "StyleSettings", ["lighting_mode", "speed", "brightness", "color_selection"]
)
_ColorSettings = namedtuple("ColorSettings", ["color1", "color2"])
_FanConfig = namedtuple(
    "FanConfig", ["mode", "duty0", "duty1", "duty2", "duty3", "duty4", "duty5", "duty6"]
)
_FanTempConfig = namedtuple(
    "FanTempConfig", ["mode", "temp0", "temp1", "temp2", "temp3", "temp4", "temp5", "temp6"]
)


@unique
class _OLEDHardwareMonitorOffset(Enum):
    CPU_FREQ = 0
    CPU_TEMP = 1
    GPU_MEMORY_FREQ = 2
    GPU_USAGE = 3
    FAN_PUMP = 4
    FAN_RADIATOR = 5
    FAN_CPUMOS = 6
    MAXIMUM = 7


@unique
class _FanMode(RelaxedNamesEnum):
    SILENT = 0
    BALANCE = 1
    GAME = 2
    CUSTOMIZE = 3
    DEFAULT = 4
    SMART = 5

    @classmethod
    def _missing_(cls, value):
        _LOGGER.debug("falling back to BALANCE for _FanMode(%s)", value)
        return _FanMode.BALANCE


@unique
class _StripeOrFan(Enum):
    STRIPE = 0
    FAN = 1


@unique
class _FanType(Enum):
    SP = 0
    HD = 1
    LL = 2


class _LEDArea(Enum):
    JCORSAIR = 53
    JCORSAIR_OUTER_LL120 = 0x40
    JPIPE1 = 11
    JPIPE2 = 21
    JRAINBOW1 = 0x1F
    JRAINBOW2 = 42
    JRGB1 = 1
    JRGB2 = 174
    ONBOARD_LED_0 = 74
    ONBOARD_LED_1 = 84
    ONBOARD_LED_10 = 174
    ONBOARD_LED_2 = 94
    ONBOARD_LED_3 = 104
    ONBOARD_LED_4 = 114
    ONBOARD_LED_5 = 124
    ONBOARD_LED_6 = 134
    ONBOARD_LED_7 = 144
    ONBOARD_LED_8 = 154
    ONBOARD_LED_9 = 164


_CYCLE_NUMBER_LED_AREA_MAPPING = {
    _LEDArea.JPIPE1.value: 20,
    _LEDArea.JPIPE2.value: 30,
    _LEDArea.JRAINBOW1.value: 41,
    _LEDArea.JRAINBOW2.value: 52,
    _LEDArea.JCORSAIR.value: 63,
}


@unique
class _LightingMode(RelaxedNamesEnum):
    BLINK = 19
    BREATHING = 2
    CLOCK = 20
    COLOR_PULSE = 21
    COLOR_RING = 15
    COLOR_RING_DOUBLE_FLASHING = 35
    COLOR_RING_FLASHING = 34
    COLOR_SHIFT = 22
    COLOR_WAVE = 23
    CORSAIR_IQUE = 37
    DISABLE = 0
    DISABLE2 = 33
    DOUBLE_FLASHING = 4
    DOUBLE_METEOR = 17
    ENERGY = 18
    FAN_CONTROL = 32
    FIRE = 38
    FLASHING = 3
    JAZZ = 12
    JRAINBOW = 28
    LAVA = 39
    LIGHTNING = 5
    MARQUEE = 24
    METEOR = 7
    MOVIE = 14
    MSI_MARQUEE = 6
    MSI_RAINBOW = 9
    NO_ANIMATION = 1
    PLANETARY = 16
    PLAY = 13
    POP = 10
    RAINBOW = 25
    RAINBOW_DOUBLE_FLASHING = 30
    RAINBOW_FLASHING = 29
    RAINBOW_WAVE = 26
    RANDOM = 31
    RAP = 11
    STACK = 36
    VISOR = 27
    WATER_DROP = 8
    END = 40


@unique
class _Speed(RelaxedNamesEnum):
    LOW = 0
    MEDIUM = 1
    HIGH = 2


@unique
class _ColorSelection(Enum):
    RAINBOW_COLOR = 0
    USER_DEFINED = 1


@unique
class _JType(Enum):
    JRAINBOW = 3
    JCORSAIR = 4
    JONBOARD = 5


@unique
class _UploadType(Enum):
    BANNER = 0
    GIF = 1


@unique
class _ScreenMode(Enum):
    HARDWARE = 0
    IMAGE = 1
    BANNER = 3
    CLOCK = 4
    SETTINGS = 5
    DISABLED = 6


class MpgCooler(UsbHidDriver):
    _COLOR_MODES = {
        "blink": _LightingMode.BLINK,
        "breathing": _LightingMode.BREATHING,
        "clock": _LightingMode.CLOCK,
        "color pulse": _LightingMode.COLOR_PULSE,
        "color ring": _LightingMode.COLOR_RING,
        "color ring double flashing": _LightingMode.COLOR_RING_DOUBLE_FLASHING,
        "color ring flashing": _LightingMode.COLOR_RING_FLASHING,
        "color shift": _LightingMode.COLOR_SHIFT,
        "color wave": _LightingMode.COLOR_WAVE,
        "corsair ique": _LightingMode.CORSAIR_IQUE,
        "disable": _LightingMode.DISABLE,
        "disable2": _LightingMode.DISABLE2,
        "double flashing": _LightingMode.DOUBLE_FLASHING,
        "double meteor": _LightingMode.DOUBLE_METEOR,
        "energy": _LightingMode.ENERGY,
        "fan control": _LightingMode.FAN_CONTROL,
        "fire": _LightingMode.FIRE,
        "flashing": _LightingMode.FLASHING,
        "jazz": _LightingMode.JAZZ,
        "jrainbow": _LightingMode.JRAINBOW,
        "lava": _LightingMode.LAVA,
        "lightning": _LightingMode.LIGHTNING,
        "marquee": _LightingMode.MARQUEE,
        "meteor": _LightingMode.METEOR,
        "movie": _LightingMode.MOVIE,
        "msi marquee": _LightingMode.MSI_MARQUEE,
        "msi rainbow": _LightingMode.MSI_RAINBOW,
        "steady": _LightingMode.NO_ANIMATION,
        "planetary": _LightingMode.PLANETARY,
        "play": _LightingMode.PLAY,
        "pop": _LightingMode.POP,
        "rainbow": _LightingMode.RAINBOW,
        "rainbow double flashing": _LightingMode.RAINBOW_DOUBLE_FLASHING,
        "rainbow flashing": _LightingMode.RAINBOW_FLASHING,
        "rainbow wave": _LightingMode.RAINBOW_WAVE,
        "random": _LightingMode.RANDOM,
        "rap": _LightingMode.RAP,
        "stack": _LightingMode.STACK,
        "visor": _LightingMode.VISOR,
        "water drop": _LightingMode.WATER_DROP,
        "end": _LightingMode.END,
    }
    BUILTIN_MODES = {
        "silent": _FanMode.SILENT.value,
        "balanced": _FanMode.BALANCE.value,
        "game": _FanMode.GAME.value,
        "default": _FanMode.DEFAULT.value,
        "smart": _FanMode.SMART.value,
    }
    SCREEN_MODES = {
        "hardware": _ScreenMode.HARDWARE,
        "image": _ScreenMode.IMAGE,
        "banner": _ScreenMode.BANNER,
        "clock": _ScreenMode.CLOCK,
        "settings": _ScreenMode.SETTINGS,
        "disable": _ScreenMode.DISABLED,
    }
    HWMONITORDISPLAY = {
        "cpu_freq": _OLEDHardwareMonitorOffset.CPU_FREQ,
        "cpu_temp": _OLEDHardwareMonitorOffset.CPU_TEMP,
        "gpu_freq": _OLEDHardwareMonitorOffset.GPU_MEMORY_FREQ,
        "gpu_usage": _OLEDHardwareMonitorOffset.GPU_USAGE,
        "fan_pump": _OLEDHardwareMonitorOffset.FAN_PUMP,
        "fan_radiator": _OLEDHardwareMonitorOffset.FAN_RADIATOR,
        "fan_cpumos": _OLEDHardwareMonitorOffset.FAN_CPUMOS,
    }
    _MATCHES = [
        (
            0x0DB0,
            0xB130,
            "MSI MPG Coreliquid K360",
            {"fan_count": 5},
        ),
        (
            0x0DB0,
            0xCA00,
            "Suspected MSI MPG Coreliquid",
            {"_unsafe": ["experimental_coreliquid_cooler"]},
        ),
        (
            0x0DB0,
            0xCA02,
            "Suspected MSI MPG Coreliquid",
            {"_unsafe": ["experimental_coreliquid_cooler"]},
        ),
    ]
    HAS_AUTOCONTROL = True

    def __init__(self, device, description, _unsafe=[], **kwargs):
        super().__init__(device, description, **kwargs)
        self._UNSAFE = _unsafe
        self._feature_data_per_led = bytearray(_PER_LED_LENGTH + 5)
        self._bytearray_oled_hardware_monitor_data = bytearray(_REPORT_LENGTH)
        self._per_led_rgb_jonboard = bytearray(_PER_LED_LENGTH)
        self._per_led_rgb_jrainbow1 = bytearray(_PER_LED_LENGTH)
        self._per_led_rgb_jrainbow2 = bytearray(_PER_LED_LENGTH)
        self._per_led_rgb_jcorsair = bytearray(_PER_LED_LENGTH)
        self._fan_count = kwargs.pop("fan_count", 5)
        # the following fields are only initialized in connect()
        self._data = None
        self._feature_data = None

    @classmethod
    def probe(cls, handle, **kwargs):
        """Probe `handle` and yield corresponding driver instances.

        These devices have multiple top-level HID usages, and HidapiDevice
        handles matching other usages have to be ignored.
        """

        # if usage_page/usage are not available due to hidapi limitations
        # (version, platform or backend), they are unfortunately left
        # uninitialized; because of this, we explicitly exclude the undesired
        # usage_page, and assume in all other cases that we either
        # have the desired usage page, or that on that system a
        # single handle is returned for that device interface (see: #259)

        if handle.hidinfo["usage_page"] == EXTRA_USAGE_PAGE:
            return
        yield from super().probe(handle, **kwargs)

    def connect(self, **kwargs):
        check_unsafe(*self._UNSAFE, error=True, **kwargs)

        ret = super().connect(**kwargs)
        self._data = kwargs.pop(
            "runtime_storage",
            RuntimeStorage(
                key_prefixes=[
                    f"vid{self.vendor_id:04x}_pid{self.product_id:04x}",
                    f"serial{self.serial_number}",
                ]
            ),
        )
        self._feature_data = self.device.get_feature_report(0x52, _MAX_DATA_LENGTH)
        self._fan_cfg = self.get_fan_config()
        self._fan_temp_cfg = self.get_fan_temp_config()
        aprom_hi, aprom_lo = self.get_firmware_version_aprom()
        self._aprom_firmware_version = aprom_hi << 4 + aprom_lo
        ldrom_hi, ldrom_lo = self.get_firmware_version_ldrom()
        self._ldrom_firmware_version = ldrom_hi << 4 + ldrom_lo
        self._oled_firmware_version = self.get_oled_firmware_version()

        return ret

    def initialize(self, **kwargs):
        check_unsafe(*self._UNSAFE, error=True, **kwargs)

        pump_mode = kwargs.pop("pump_mode", "balanced")
        direction = kwargs.pop("direction", "default")
        if pump_mode == "balanced":
            pump_mode = "balance"
        pump_mode_int = _FanMode[pump_mode].value
        self._data.store("pump_mode", pump_mode_int)
        dir_int = 0
        if direction not in ("default", "top", "bottom", "left", "right", "0", "1", "2", "3"):
            _LOGGER.warning(
                "unknown direction value: correct values are 0-3 or top, "
                "bottom, left, right, default."
            )
        if direction in ("1", "left"):
            dir_int = 1
        elif direction in ("2", "bottom"):
            dir_int = 2
        elif direction in ("3", "right"):
            dir_int = 3
        self._data.store("direction", dir_int)
        self.set_oled_brightness_and_direction(100, dir_int)
        if pump_mode_int == _FanMode.GAME.value:
            self.switch_to_game_mode()
        elif pump_mode_int == _FanMode.BALANCE.value:
            self.switch_to_balance_mode()
        elif pump_mode_int == _FanMode.DEFAULT.value:
            self.switch_to_default_mode()
        elif pump_mode_int == _FanMode.SILENT.value:
            self.switch_to_silent_mode()
        elif pump_mode_int == _FanMode.SMART.value:
            self.switch_to_smart_mode()

        return [
            ("Display firmware version", self._oled_firmware_version, ""),
            ("APROM firmware version", self._aprom_firmware_version, ""),
            ("LDROM firmware version", self._ldrom_firmware_version, ""),
            ("Serial number", self.serial_number, ""),
            ("Pump mode", pump_mode, ""),
        ]

    def get_status(self, **kwargs):
        if not check_unsafe(*self._UNSAFE, **kwargs):
            _LOGGER.warning(
                f"{self.description}: disabled, requires unsafe features "
                f"'{','.join(self._UNSAFE)}'"
            )
            return []
        self._write((0x31,))
        array = self._read()
        assert array[1] == 0x31, "Unexpected value in response buffer"
        return [
            ("Fan 1 speed", u16le_from(array, offset=2), "rpm"),
            ("Fan 1 duty", u16le_from(array, offset=0x16), "%"),
            ("Fan 2 speed", u16le_from(array, offset=4), "rpm"),
            ("Fan 2 duty", u16le_from(array, offset=0x18), "%"),
            ("Fan 3 speed", u16le_from(array, offset=6), "rpm"),
            ("Fan 3 duty", u16le_from(array, offset=0x1A), "%"),
            ("Water block speed", u16le_from(array, offset=8), "rpm"),
            ("Water block duty", u16le_from(array, offset=0x1C), "%"),
            ("Pump speed", u16le_from(array, offset=0xA), "rpm"),
            ("Pump duty", u16le_from(array, offset=0x1E), "%"),
            # Temperature values are not used by the K360 model, it only reports
            # some default values that are not meaningful for the user.
            # https://github.com/liquidctl/liquidctl/pull/564#discussion_r1450753883
            # ('Temperature inlet', u16le_from(array, offset=12), '°C'),
            # ('Temperature outlet', u16le_from(array, offset=14), '°C'),
            # ('Temperature sensor 1', u16le_from(array, offset=16), '°C'),
            # ('Temperature sensor 2', u16le_from(array, offset=18), '°C'),
        ]

    def set_time(self, time, **kwargs):
        check_unsafe(*self._UNSAFE, error=True, **kwargs)
        return self.set_oled_clock(time)

    def set_hardware_status(self, T, cpu_f=0, gpu_f=0, gpu_U=0, **kwargs):
        check_unsafe(*self._UNSAFE, error=True, **kwargs)
        self.set_oled_show_cpu_status(cpu_f, T)
        self.set_oled_gpu_status(gpu_f, gpu_U)

    def get_fan_config(self):
        self._write((0x32,))
        buf = self._read()
        assert buf[1] == 0x32, "Unexpected value in returned list"
        ret = []
        for mode_index in (2, 10, 18, 26, 34):
            ret.append(_FanConfig(*buf[mode_index : mode_index + 8]))
        return ret

    def set_fan_config(self, configs):
        buf = bytearray(_REPORT_LENGTH - 1)
        buf[0] = 0x40
        for config, offset in zip(configs, (1, 9, 17, 25, 33)):
            buf[offset] = config.mode
            buf[offset + 1] = config.duty0
            buf[offset + 2] = config.duty1
            buf[offset + 3] = config.duty2
            buf[offset + 4] = config.duty3
            buf[offset + 5] = config.duty4
            buf[offset + 6] = config.duty5
            buf[offset + 7] = config.duty6
        return self._write(buf)

    def get_fan_temp_config(self):
        self._write((0x33,))
        buf = self._read()
        assert buf[1] == 0x32, "Unexpected value in returned list"
        ret = []
        for mode_index in (2, 10, 18, 26, 34):
            ret.append(_FanTempConfig(*buf[mode_index : mode_index + 8]))
        return ret

    def set_fan_temp_config(self, configs):
        buf = bytearray(_REPORT_LENGTH)
        buf[0] = 0x41
        for config, offset in zip(configs, (1, 9, 17, 25, 33)):
            buf[offset] = config.mode
            buf[offset + 1] = config.temp0
            buf[offset + 2] = config.temp1
            buf[offset + 3] = config.temp2
            buf[offset + 4] = config.temp3
            buf[offset + 5] = config.temp4
            buf[offset + 6] = config.temp5
            buf[offset + 7] = config.temp6
        return self._write(buf)

    def _send_safe_temp(self):
        _LOGGER.info(
            "duty profiles on this device require continuous communication, "
            "setting initial control temperature to 100C for safety."
        )
        self.set_oled_show_cpu_status(0, 100)

    def set_profiles(self, channels, profiles, **kwargs):
        """
        Set custom or device preset fan curve for multiple channels.

        NOTE: The device will not keep updating its duty cycles
        automatically after this function is called. The device
        manages duties according to the previous temperature
        sent to it via device.set_oled_show_cpu_status()
        """
        check_unsafe(*self._UNSAFE, error=True, **kwargs)

        fan_cfg = self.get_fan_config()
        fan_temp_cfg = self.get_fan_temp_config()
        channel_idx = [self.parse_channel(ch) for ch in channels]

        for idx, prof in zip(channel_idx, profiles):
            if type(prof) == str:
                fanmode = _FanConfig(self.BUILTIN_MODES[prof], 0, 0, 0, 0, 0, 0, 0)
                tempmode = _FanTempConfig(self.BUILTIN_MODES[prof], 0, 0, 0, 0, 0, 0, 0)
            else:
                duties, temps = map(self.clamp_and_pad, zip(*prof))
                fanmode = _FanConfig(_FanMode.CUSTOMIZE.value, *duties)
                tempmode = _FanTempConfig(_FanMode.CUSTOMIZE.value, *temps)
            for i in idx:
                fan_cfg[i] = fanmode
                fan_temp_cfg[i] = tempmode

        self._fan_cfg = fan_cfg
        self._fan_temp_cfg = fan_temp_cfg

        self.set_fan_config(fan_cfg)
        self.set_fan_temp_config(fan_temp_cfg)
        self._send_safe_temp()

    def parse_channel(self, channel):
        if channel == "pump":
            return [4]
        elif channel == "fans":
            return range(_RAD_FAN_COUNT)
        elif channel == "waterblock-fan":
            return [3]
        elif channel[:3] == "fan" and (int(channel[3:]) in range(_RAD_FAN_COUNT)):
            return [int(channel[3:])]
        else:
            raise ValueError(
                'unknown channel, should be "fans", "fan1", "fan2", "fan3", "waterblock-fan" or "pump".'
            )

    @staticmethod
    def clamp_and_pad(values):
        return ([clamp(v, 0, 100) for v in values] + [0] * _MAX_DUTIES)[:_MAX_DUTIES]

    def set_speed_profile(self, channel, profile, **kwargs):
        """
        Set custom fan curve for a given channel.

        NOTE: The device will not keep updating its duty cycles
        automatically after this function is called. The device
        manages duties according to the previous temperature
        sent to it via device.set_oled_show_cpu_status()
        """

        check_unsafe(*self._UNSAFE, error=True, **kwargs)

        duties_temps = list(zip(*profile))
        duties, temps = tuple(self.clamp_and_pad(v) for v in duties_temps)

        for i in self.parse_channel(channel):
            self._fan_cfg[i] = _FanConfig(_FanMode.CUSTOMIZE.value, *duties)
            self._fan_temp_cfg[i] = _FanTempConfig(_FanMode.CUSTOMIZE.value, *temps)

        self.set_fan_config(self._fan_cfg)
        self.set_fan_temp_config(self._fan_temp_cfg)
        self._send_safe_temp()

    def set_fixed_speed(self, channel, duty, **kwargs):
        channel_nums = self.parse_channel(channel)

        check_unsafe(*self._UNSAFE, error=True, **kwargs)

        for i in channel_nums:
            self._fan_cfg[i] = _FanConfig(_FanMode.CUSTOMIZE.value, *([duty] * _MAX_DUTIES))
            self._fan_temp_cfg[i] = _FanTempConfig(_FanMode.CUSTOMIZE.value, 0, 0, 0, 0, 0, 0, 0)

        self.set_fan_config(self._fan_cfg)
        self.set_fan_temp_config(self._fan_temp_cfg)

    def set_color(self, channel, mode, colors, speed=1, brightness=10, color_selection=1, **kwargs):
        check_unsafe(*self._UNSAFE, error=True, **kwargs)

        assert channel == "sync", f'Unexpected lighting channel {channel}. Supported: "sync"'
        colors = list(colors)
        if not colors:
            color_selection = 0
        else:
            if len(colors) == 1:
                colors.append((0, 0, 0))
            self.set_color_setting(_LEDArea.JRAINBOW1.value, *colors[0], *colors[1])

        mode = self._COLOR_MODES[mode].value
        self.set_style_setting(
            _LEDArea.JRAINBOW1.value, mode, int(speed), brightness, color_selection
        )
        self.set_send_led_setting(1)

    def get_current_model_index(self):
        self._write((0xB1,), 0xCC, prefix=1)
        return self._read()[2]

    def set_screen(self, channel, mode, value, **kwargs):
        check_unsafe(*self._UNSAFE, error=True, **kwargs)

        assert channel.lower() == "lcd"
        try:
            mode = self.SCREEN_MODES[mode]
        except KeyError as e:
            raise Exception(
                f"Unknown screen mode! Should be one of: {self.SCREEN_MODES.keys()}"
            ) from e

        if value is not None:
            opts = value.split(";")

        if mode == _ScreenMode.HARDWARE:
            # hardware monitor options are a list of the values to display
            # (case-insensitive keys to MPGCooler.HWMONITORDISPLAY)
            self.set_oled_show_hardware_monitor(opts)

        if mode == _ScreenMode.IMAGE:
            # image options are: image-type, image-index[, image-file]
            if len(opts) == 3:
                self.set_oled_upload_gif(opts)
            elif len(opts) == 2:
                self.set_oled_show_profile(opts)
            else:
                raise ValueError(
                    f"Unexpected options for LCD image. Expected either:"
                    '"image-type;image-slot" or '
                    '"image-type;image-slot;image-file", '
                    "instead got: {opts}"
                )

        elif mode == _ScreenMode.BANNER:
            # banner options are: banner-type, banner-index, message[, image-file]
            if (len(opts) == 3) or (len(opts) == 4):
                if len(opts) == 4:
                    save_slot = int(opts[1])
                    assert save_slot >= 4, (
                        "Cannot overwrite preset banner images, "
                        "please use save slots starting from 4 for your uploaded files"
                    )
                    img = self._prepare_bmp(opts[3])
                    self.set_oled_upload_banner(img, banner_no=save_slot)
                self.set_oled_user_message(opts[2])
                self.set_oled_show_banner(banner_type=int(opts[0]), bmp_no=int(opts[1]))
            else:
                raise ValueError(
                    f"Unexpected options for LCD banner. Expected either:"
                    '"banner-type;save-slot;message" or '
                    '"banner-type;save-slot;message;image-file", '
                    "instead got: {opts}"
                )

        elif mode == _ScreenMode.CLOCK:
            # clock option is the display style
            self.set_oled_show_clock(int(opts[0]))

        elif mode == _ScreenMode.SETTINGS:
            # setting options are: brightness, direction
            brightness, direction = [int(x) for x in opts]
            self.set_oled_brightness_and_direction(brightness=brightness, direction=direction)

        elif mode == _ScreenMode.DISABLED:
            # switches off the display
            self.set_oled_show_disable()

    def _prepare_bmp(self, path):
        end_w, end_h = 240, 320
        img = Image.open(path)
        w, h = img.size
        wrat = end_w / w
        hrat = end_h / h
        ratio = wrat if wrat > hrat else hrat
        img = img.resize((int(ratio * w), int(ratio * h)))
        w, h = img.size
        x_start = int((w - end_w) / 2)
        y_start = int((h - end_h) / 2)
        img = img.crop((x_start, y_start, x_start + end_w, y_start + end_h))

        img = img.convert("RGB")
        img_bytes = io.BytesIO()
        img.save(img_bytes, format="BMP")
        return img_bytes

    def get_firmware_version_aprom(self):
        self._write((0xB0,), 0xCC, prefix=1)
        ret = self._read()
        return (
            ret[2] >> 4,  # high
            ret[2] & 0xF,  # low
        )

    def get_firmware_version_ldrom(self):
        self._write((0xB6,), 0xCC, prefix=1)
        ret = self._read()
        return (
            ret[2] >> 4,  # high
            ret[2] & 0xF,  # low
        )

    def get_firmware_checksum_aprom(self):
        self._write((0xB0,), 0xCC, prefix=1)
        ret = self._read()
        return (
            ret[8],  # high
            ret[9],  # low
        )

    def get_firmware_checksum_ldrom(self):
        self._write((0xB4,), 0xCC, prefix=1)
        ret = self._read()
        return (
            ret[8],  # high
            ret[9],  # low
        )

    def get_all_hardware_monitor(self):
        self.device.clear_enqueued_reports()
        return self.device.get_feature_report(0xD0, _REPORT_LENGTH)

    def set_buzzer(self, type, frequency=650):
        return self._write(
            (0xC2, 0, 0, 0, 0, 0, type, frequency & 0xFF, (frequency >> 8) & 0xFF), prefix=1
        )

    def set_led_global(self, on_off):
        return self._write((0xBB, 0, 0, 0, 0, int(on_off)), prefix=1)

    def get_led_global(self):
        self._write((0xBA,), 0xCC, prefix=1)
        ret = self._read()
        ok = len(ret) == _REPORT_LENGTH
        for j, _ in enumerate(ret):
            if (
                (j == 0 and ret[j] != 1)
                or (j == 1 and ret[j] != 90)
                or (j == 6 and ret[j] not in (0, 1))
            ):
                ok = False
            elif ret[j] != 0xCC:
                ok = False
        return ok and ret[6] == 1

    def get_led_pe0(self):
        self._write((0xA0, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 1, 0xF2), 0xCC, prefix=0xFA)
        ret = self._read()
        ok = True
        if (
            ret[0] != 250
            or ret[1] != 160
            or ret[2] != 205
            or ret[9] != 1
            or ret[10] != 242
            or ret[11] not in (0, 1)
        ):
            ok = 0
        for j, n in enumerate(ret, start=23):
            if j > 29:
                break
            if n != 0:
                ok = False
        for j, n in enumerate(ret, start=30):
            if n != 0xCC:
                ok = False
        return ok and ret[11] == 1

    def get_device_settings(self):
        return _DeviceSettings(
            self._feature_data[61] & 1,
            (self._feature_data[61] & 0xE) >> 1,
            (self._feature_data[62] & 0xFC) >> 2,
            self._feature_data[72] & 1,
            self._feature_data[41],
            self._feature_data[52],
            self._feature_data[63],
        )

    def get_board_sync_settings(self):
        return _BoardSyncSettings(
            (self._feature_data[82] & 1) == 1,
            (self._feature_data[78] & 0x80) >> 7 == 1,
            ((self._feature_data[82] >> 4) & 1) == 1,
            ((self._feature_data[82] >> 5) & 1) == 1,
            ((self._feature_data[82] >> 1) & 1) == 1,
            ((self._feature_data[82] >> 2) & 1) == 1,
            ((self._feature_data[82] >> 3) & 1) == 1,
        )

    def get_style_settings(self, led_area):
        return _StyleSettings(
            self._feature_data[led_area],
            (self._feature_data[led_area + 4] & 3),
            (self._feature_data[led_area + 4] >> 2) & 0x1F,
            ((self._feature_data[led_area + 8] & 0x80) >> 7),
        )

    def get_color_settings(self, led_area):
        return _ColorSettings(
            self._feature_data[led_area + 1 : led_area + 4],
            self._feature_data[led_area + 5 : led_area + 8],
        )

    def get_current_led_setting(self):
        self._feature_data = self.device.get_feature_report(self._feature_data[0], _MAX_DATA_LENGTH)
        self._feature_data[62] = clamp(self._feature_data[62] - 4, 0, 255)
        return self._feature_data

    def get_all_board(self):
        return self.device.get_feature_report(0x52, _MAX_DATA_LENGTH)

    def get_oled_firmware_version(self):
        self._write((0xF1,))
        return self._read()[2]

    def get_oled_gif_checksum(self):
        self._write((0xC2,))
        ret = self._read()
        return (
            ret[3],  # high
            ret[2],  # low
        )

    def get_oled_banner_checksum(self):
        self._write((0xD2,))
        buf = self._read()
        return (
            buf[3],  # high
            buf[2],  # low
        )

    def get_oled_m481checksum(self):
        self._write((0xF1,))
        ret = self._read()
        return (
            ret[3],  # high
            ret[2],  # low
        )

    def set_volume(self, main, left, right):
        return self._write(
            (0xC0, clamp(main, 0, 100), clamp(left, 0, 100), clamp(right, 0, 100)), 0xCC, prefix=1
        )

    def set_oled_cpu_message(self, message):
        return self._write([0x90] + list(message[:60].encode("ascii", "ignore")))

    def set_oled_show_hardware_monitor(self, opts, radiator_fan_smart_mode_on_off=True):
        """
        Command device to display hardware information on the built in display.

        Parameters
        ----------
        opts:
            Array[str] Indicates which hardware features will be presented on the screen,
                       from CPU_FREQ, CPU_TEMP, GPU_MEMORY_FREQ, GPU_USAGE, FAN_PUMP, FAN_RADIATOR, FAN_CPUMOS
        radiator_fan_smart_mode_on_off:
            Bool Indicates whether smart control is enabled, this slightly alters the visualization
        """
        show_idx = [self.HWMONITORDISPLAY[x.lower()].value for x in opts]
        show_area = [i in show_idx for i in range(_OLEDHardwareMonitorOffset.MAXIMUM.value + 1)]
        if len(show_area) < 7:
            return False
        buf = bytearray(_REPORT_LENGTH)
        buf[0] = 0xD0
        buf[1] = 0x71
        for item in iter(_OLEDHardwareMonitorOffset):
            if item.value != _OLEDHardwareMonitorOffset.MAXIMUM and show_area[item.value]:
                buf[item.value + 2] = 1
            if show_area[5]:
                buf[9] = 3 if radiator_fan_smart_mode_on_off else 1
        return self._write(buf[1:])

    def set_return_to_default(self):
        """--reset-all"""
        self._feature_data = copy(_DEFAULT_FEATURE_DATA)
        self._feature_data[184] = 1
        return self._set_all_board(self._feature_data)

    def _set_all_board(self, data):
        if data[41] > 200:
            data[41] = 100
        if data[52] > 240:
            data[52] = 100
        b = (data[62] & 0xFC) >> 2
        if ((b + 1) * data[63]) > 240:
            b = 5
            data[62] = b << 2
            data[63] = 12
        if len(data) < 185 or data[0] != 82:
            return False
        return bool(self.device.send_feature_report(data))

    def set_cycle_number(self, stripe_type, cycle_num):
        """Cycle number should NOT be clamped."""
        if stripe_type in _CYCLE_NUMBER_STRIPE_TYPE_MAPPING:
            self._feature_data[_CYCLE_NUMBER_STRIPE_TYPE_MAPPING[stripe_type]] = cycle_num

    def set_cycle_number_by_led_area(self, led_area, cycle_num):
        if led_area in _CYCLE_NUMBER_LED_AREA_MAPPING:
            self._feature_data[_CYCLE_NUMBER_LED_AREA_MAPPING[led_area]] = cycle_num

    def set_device_setting(
        self, stripe_or_fan, fan_type, corsair_device_qty, ll120_outer_individual
    ):
        corsair_device_qty = clamp(corsair_device_qty, 0, 63)
        ll120_outer_individual = clamp(int(ll120_outer_individual), 0, 1)
        self._feature_data[61] = (self._feature_data[61] & 0x80) | (fan_type << 1) | stripe_or_fan
        self._feature_data[62] = corsair_device_qty << 2
        self._feature_data[72] = self._feature_data[72] | ll120_outer_individual

    def set_board_sync_setting(
        self,
        onboard_sync,
        combine_jrgb,
        combine_jpipe1,
        combine_jpipe2,
        combine_jrainbow1,
        combine_jrainbow2,
        combine_jcorsair,
    ):
        self._feature_data[82] |= clamp(int(onboard_sync), 0, 1)
        self._feature_data[78] |= 0b10000000 if combine_jrgb else 0
        self._feature_data[82] |= 0b00010000 if combine_jpipe1 else 0
        self._feature_data[82] |= 0b00100000 if combine_jpipe2 else 0
        self._feature_data[82] |= 0b00000010 if combine_jrainbow1 else 0
        self._feature_data[82] |= 0b00000100 if combine_jrainbow2 else 0
        self._feature_data[82] |= 0b00001000 if combine_jcorsair else 0

    def set_style_setting(self, led_area, lighting_mode, speed, brightness, color_selection):
        """
        --led-area
        --lighting-mode
        --speed
        --brightness
        --color-selection
        """
        lighting_mode = clamp(lighting_mode, 0, 40)
        speed = clamp(speed, 0, 2)
        brightness = clamp(brightness, 0, 10)
        color_selection = clamp(color_selection, 0, 1)
        self._feature_data[led_area] = lighting_mode
        self._feature_data[led_area + 4] = (
            (self._feature_data[led_area + 4] & 0x80) | (brightness << 2) | speed
        )
        self._feature_data[led_area + 8] = 0b10000000 if color_selection else 0

    def set_color_setting(
        self, led_area, color1_r, color1_g, color1_b, color2_r, color2_g, color2_b
    ):
        """
        --led-area
        --color1
        --color2
        """

        self._feature_data[led_area + 1] = clamp(color1_r, 0, 255)
        self._feature_data[led_area + 2] = clamp(color1_g, 0, 255)
        self._feature_data[led_area + 3] = clamp(color1_b, 0, 255)
        self._feature_data[led_area + 4] = clamp(led_area, 0, 174)
        self._feature_data[led_area + 5] = clamp(color2_r, 0, 255)
        self._feature_data[led_area + 6] = clamp(color2_g, 0, 255)
        self._feature_data[led_area + 7] = clamp(color2_b, 0, 255)

    def set_send_led_setting(self, save):
        """applies changes, persists to device with save option"""
        self._feature_data[184] = int(bool(save))
        return self._set_all_board(self._feature_data)

    def set_direction_setting_b931_only(
        self, board_sync, jrainbow1, jrainbow2, jrainbow3, jrainbow4, jrainbow5
    ):
        """--b931-direction ?"""
        self._feature_data[83] |= int(bool(board_sync))
        self._feature_data[39] |= int(bool(jrainbow1))
        self._feature_data[61] |= int(bool(jrainbow2))
        self._feature_data[50] |= int(bool(jrainbow3))
        self._feature_data[19] |= int(bool(jrainbow4))
        self._feature_data[29] |= int(bool(jrainbow5))

    def set_oled_user_message(self, message):
        """--message"""
        return self._write([0x93] + list((message[:61] + " ").encode("ascii", "ignore")))

    def set_oled_show_gameboot(self, selection, message):
        """
        --selection
        --message
        """
        return self._write([0x73, selection] + list(message[:60].encode("ascii", "ignore")))

    def set_oled_show_profile(self, opts):
        """
        --profile-type
        --gif-number
        """
        profile_type, gif_no = map(int, opts[:2])
        data = [0x70, 0, gif_no]
        self._write(data)
        data[1] = clamp(profile_type, 0, 1)
        return self._write(data)

    def set_oled_show_banner(self, banner_type=0, bmp_no=0):
        """
        --banner-type
        --bmp-number
        """
        data = [0x79, banner_type, bmp_no]
        return self._write(data)

    def set_oled_show_disable(self):
        """--disable-oled"""
        return self._write((0x7F,))

    def _set_oled_upload(self, type, bytes, type_num=0):
        is_gif = type == _UploadType.GIF
        start_cmd = 0xC0 if is_gif else 0xD0
        content = bytes.getbuffer()
        l = len(content)
        if l > (2**20):
            raise ValueError("file size of image is too large, something went wrong!")
        _LOGGER.debug(f"size of uploaded image is {l} bytes.")
        self._write(
            (start_cmd, l & 0xFF, (l >> 8) & 0xFF, (l >> 16) & 0xFF, (l >> 24) & 0xFF, type_num)
        )
        sleep(2)
        checksum = sum(content) & 0xFFFF
        n = 0
        while n < l:
            array = [start_cmd + 1]
            o = clamp(l - n, 0, 60)
            for k in range(0, o):
                array.append(content[n + k])
            n += o
            self._write(array)
        if is_gif:
            high, low = self.get_oled_gif_checksum()
            check = (high << 8) + low
        else:
            high, low = self.get_oled_banner_checksum()
            check = (high << 8) + low

        if check != checksum:
            _LOGGER.error(
                f"invalid upload, image checksums: high {high} vs {checksum & 0xFF00}, "
                f"low {low} vs {checksum & 0xFF}."
            )

    def set_oled_upload_gif(self, opts):
        """
        --upload-gif-file
        --upload-gif-number
        """
        imgtype, gif_no = map(int, opts[:2])
        assert imgtype == 1, "Cannot override default images (image type 0)"
        file = opts[2]
        image_bytes = self._prepare_bmp(file)
        self._set_oled_upload(_UploadType.GIF, image_bytes, gif_no)

    def set_oled_upload_banner(self, bytes, banner_no=4):
        """Default is 4 to not overwrite the default banners.

        --upload-banner-file
        --upload-banner-number
        """
        self._set_oled_upload(_UploadType.BANNER, bytes, banner_no)

    def set_oled_clock(self, time):
        """
        Sends the specified time to the device.
        """
        return self._write(
            (
                0x83,
                time.year % 100,
                time.month,
                time.day,
                time.weekday(),
                time.hour,
                time.minute,
                time.second,
            )
        )

    def set_oled_show_clock(self, style):
        """--set-oled-mode clock"""
        return self._write((0x7A, clamp(style, 0, 2)))

    def set_oled_show_cpu_status(self, freq, temp):
        """--update-cpu-status"""
        freq = clamp(int(freq), 0, 65536)
        temp = clamp(int(temp), 0, 65536)
        return self._write(
            (
                0x85,
                freq & 0xFF,
                (freq >> 8) & 0xFF,
                temp & 0xFF,
                (temp >> 8) & 0xFF,
            )
        )

    @staticmethod
    def _make_buffer(array, fill=0, total_size=_REPORT_LENGTH, prefix=0xD0):
        return bytearray([prefix] + list(array) + ((total_size - (len(array) + 1)) * [fill]))

    def _write(self, array, fill=0, total_size=_REPORT_LENGTH, prefix=0xD0):
        self.device.clear_enqueued_reports()
        return self.device.write(self._make_buffer(array, fill, total_size, prefix))

    def _read(self, size=_REPORT_LENGTH):
        return bytearray(self.device.read(size))

    def set_oled_gpu_status(self, mem_freq, usage):
        """
        --set-gpu-memory-frequency
        --set-gpu-usage
        """
        mem_freq = clamp(int(mem_freq), 0, 65536)
        usage = clamp(int(usage), 0, 65536)
        return self._write(
            (
                0x86,
                mem_freq & 0xFF,
                (mem_freq >> 8) & 0xFF,
                usage & 0xFF,
                (usage >> 8) & 0xFF,
            )
        )

    def set_oled_brightness_and_direction(self, brightness=100, direction=0):
        """
        --set-oled-brightness
        --direction
        """
        return self._write((0x7E, clamp(brightness, 0, 100), clamp(direction, 0, 3)))

    def set_per_led_720byte(self, jtype, area, rgb_data):
        b = jtype
        if self.product_id != 0x7B10 and self.product_id != 0x7C34:
            b += 1
        if len(rgb_data) < 3:
            rgb_data = bytearray(3)
        rgb_data = rgb_data[:_PER_LED_LENGTH]
        self._feature_data_per_led = self._make_buffer(
            [0x37, b, area] + list(rgb_data), total_size=_PER_LED_LENGTH + 5, prefix=0x53
        )
        return self.device.send_feature_report(self._feature_data_per_led)

    def set_per_led_index(self, jtype, area, index_and_rgb, show=True, led_count=0):
        """
        --set-per-led-index
        --jtype INT
        --area INT
        --index-and-rgb hex string?
        --show (or not passed)
        --maximum-leds INT
        """
        array = bytearray(1)

        def rank_check(array):
            return (
                isinstance(array, Sequence)
                and all(isinstance(x, Sequence) for x in array)
                and all(not isinstance(xa, Sequence) for x in array for xa in x)
            )

        if len(index_and_rgb[1]) < 4 or len(index_and_rgb[0]) < 1 or not rank_check(index_and_rgb):
            raise ValueError("index_and_rgb should be a 2-dimensional list")
        if jtype == _JType.JONBOARD.value:
            array = self._per_led_rgb_jonboard
        elif jtype == _JType.JRAINBOW.value:
            area = clamp(area, 0, 1)
            array = self._per_led_rgb_jrainbow2 if area == 1 else self._per_led_rgb_jrainbow1
        elif jtype == _JType.JCORSAIR.value:
            array = self._per_led_rgb_jcorsair
        end_index = led_count if led_count > 0 else len(index_and_rgb)
        for i in range(end_index):
            if index_and_rgb[i][0] * 3 + 2 < len(array):
                array[index_and_rgb[i][0] * 3] = index_and_rgb[i][1]
                array[index_and_rgb[i][0] * 3 + 1] = index_and_rgb[i][2]
                array[index_and_rgb[i][0] * 3 + 2] = index_and_rgb[i][3]
        if show:
            self.set_per_led_720byte(jtype, area, array)

    def set_switch_to_per_led_mode(self, jtype, area, show=False):
        """
        --switch-to-per-led-mode
        --jtype INT
        --area INT
        --show (or not passed)
        """
        if jtype == _JType.JONBOARD.value:
            self.set_style_setting(
                _LEDArea.ONBOARD_LED_0.value,
                _LightingMode.CORSAIR_IQUE.value,
                _Speed.MEDIUM.value,
                10,
                _ColorSelection.USER_DEFINED.value,
            )
            self._per_led_rgb_jonboard = bytearray(_PER_LED_LENGTH)
        elif jtype == _JType.JRAINBOW.value:
            if area == 0:
                self._per_led_rgb_jrainbow1 = bytearray(_PER_LED_LENGTH)
                self.set_cycle_number(0, 200)
                self.set_style_setting(
                    _LEDArea.JRAINBOW1.value,
                    _LightingMode.CORSAIR_IQUE.value,
                    _Speed.MEDIUM.value,
                    10,
                    _ColorSelection.USER_DEFINED.value,
                )
            elif area == 1:
                self._per_led_rgb_jrainbow2 = bytearray(_PER_LED_LENGTH)
                self.set_cycle_number(1, 240)
                self.set_style_setting(
                    _LEDArea.JRAINBOW2.value,
                    _LightingMode.CORSAIR_IQUE.value,
                    _Speed.MEDIUM.value,
                    10,
                    _ColorSelection.USER_DEFINED.value,
                )
        elif jtype == _JType.JCORSAIR.value:
            self._per_led_rgb_jcorsair = bytearray(_PER_LED_LENGTH)
            self.set_style_setting(
                _LEDArea.JCORSAIR.value,
                _LightingMode.CORSAIR_IQUE.value,
                _Speed.MEDIUM.value,
                10,
                _ColorSelection.USER_DEFINED.value,
            )
            if ((self._feature_data[61] & 0xE) >> 1) == 0 and (self._feature_data[61] & 1) == 1:
                self.set_device_setting(_StripeOrFan.FAN.value, _FanType.SP.value, 5, 0)
            else:
                self.set_cycle_number(2, 240)
                self.set_device_setting(_StripeOrFan.STRIPE.value, _FanType.HD.value, 0, 0)
        self.set_board_sync_setting(True, True, True, True, False, False, False)
        if show:
            self.set_send_led_setting(False)
            self.set_clear_per_led(jtype, area)

    def set_clear_per_led(self, jtype, area):
        """
        --clear-per-led
        --jtype
        --area
        """
        return self.set_per_led_720byte(jtype, area, bytearray(3))

    def _set_all(self, which, r=0, g=0, b=0):
        self.set_board_sync_setting(True, True, True, True, True, True, True)
        self.set_style_setting(_LEDArea.ONBOARD_LED_0.value, which, 1, 10, 1)
        self.set_color_setting(_LEDArea.ONBOARD_LED_0.value, r, g, b, r, g, b)
        self.set_send_led_setting(False)

    def set_all_disable(self):
        """--led-disable"""
        self._set_all(_LightingMode.DISABLE.value)

    def set_all_static(self, r, g, b):
        """--led-all-static"""
        self._set_all(_LightingMode.NO_ANIMATION.value, r, g, b)

    def set_all_flashing(self, r, g, b):
        """--led-all-flashing"""
        self._set_all(_LightingMode.FLASHING.value, r, g, b)

    def set_all_breathing(self, r, g, b):
        """--led-all-breathing"""
        self._set_all(_LightingMode.BREATHING.value, r, g, b)

    def set_all_rainbow_wave(self):
        """--led-all-rainbow-wave"""
        self._set_all(_LightingMode.RAINBOW_WAVE.value)

    def switch_to_balance_mode(self):
        self.set_fan_config(
            [_FanConfig(_FanMode.BALANCE.value, 0, 0, 0, 0, 0, 0, 0)] * self._fan_count
        )
        self.set_fan_temp_config(
            [_FanTempConfig(_FanMode.BALANCE.value, 0, 0, 0, 0, 0, 0, 0)] * self._fan_count
        )
        self._send_safe_temp()

    def switch_to_game_mode(self):
        self.set_fan_config(
            [_FanConfig(_FanMode.GAME.value, 0, 0, 0, 0, 0, 0, 0)] * self._fan_count
        )
        self.set_fan_temp_config(
            [_FanTempConfig(_FanMode.GAME.value, 0, 0, 0, 0, 0, 0, 0)] * self._fan_count
        )
        self._send_safe_temp()

    def switch_to_silent_mode(self):
        self.set_fan_config(
            [_FanConfig(_FanMode.SILENT.value, 0, 0, 0, 0, 0, 0, 0)] * self._fan_count
        )
        self.set_fan_temp_config(
            [_FanTempConfig(_FanMode.SILENT.value, 0, 0, 0, 0, 0, 0, 0)] * self._fan_count
        )
        self._send_safe_temp()

    def switch_to_smart_mode(self):
        self.set_fan_config(
            [_FanConfig(_FanMode.SMART.value, 0, 0, 0, 0, 0, 0, 0)] * self._fan_count
        )
        self.set_fan_temp_config(
            [_FanTempConfig(_FanMode.SMART.value, 0, 0, 0, 0, 0, 0, 0)] * self._fan_count
        )
        self._send_safe_temp()

    def switch_to_default_mode(self):
        self.set_fan_config(
            [_FanConfig(_FanMode.DEFAULT.value, 0, 0, 0, 0, 0, 0, 0)] * self._fan_count
        )
        self.set_fan_temp_config(
            [_FanTempConfig(_FanMode.DEFAULT.value, 0, 0, 0, 0, 0, 0, 0)] * self._fan_count
        )
        self._send_safe_temp()

    # def set_oled_cpu_message(self, message):
    #     return self._write([0x90] +
    #                        list(message[:60].encode('ascii', 'ignore')))

    # def set_oled_memory_message(self, message):
    #     return self._write([0x91] +
    #                        list(message[:60].encode('ascii', 'ignore')))

    # def set_oled_vga_message(self, message):
    #     return self._write([0x92] +
    #                        list(message[:60].encode('ascii', 'ignore')))

    # def set_oled_show_system_message(self):
    #     return self._write((0x72, ))

    # def set_oled_show_user_message(self):
    #     return self._write((0x74, ))

    # def set_oled_start_isp_process(self):
    #     return self._write((0xfa, ))

    # def set_oled_show_demo_mode(self):
    #     return self._write((0x77, 0xff))

    # def set_reset_mcu(self):
    #     return self._write((0xd0, ), 0xcc, prefix=1)



# uses the psf/black style

from struct import pack
from datetime import datetime

import pytest
from _testutils import MockHidapiDevice, Report

from liquidctl.driver.msi import MpgCooler, _REPORT_LENGTH, _DEFAULT_FEATURE_DATA, _LightingMode
from liquidctl.error import UnsafeFeaturesNotEnabled


@pytest.fixture
def mpgCoreLiquidK360Device():
    description = "Mock MPG CoreLiquid K360"
    device = _MockCoreLiquid(vendor_id=0xFFFF, product_id=0xB130)
    dev = MpgCooler(device, description)

    dev.connect()
    return dev


@pytest.fixture
def mpgCoreLiquidDeviceExperimental():
    _, pid, desc, kwargs = MpgCooler._MATCHES[-1]
    description = "Mock " + desc
    unsafe = kwargs["_unsafe"]
    device = _MockCoreLiquid(vendor_id=0xFFFF, product_id=pid)
    dev = MpgCooler(device, description, **kwargs, unsafe=unsafe)

    dev.connect(unsafe=unsafe)
    return dev


@pytest.fixture
def mpgCoreLiquidK360DeviceInvalid():
    description = "Mock MPG CoreLiquid K360"
    device = _MockCoreLiquidInvalid(vendor_id=0xFFFF, product_id=0xB130)
    dev = MpgCooler(device, description)

    dev.connect()
    return dev


class _MockCoreLiquid(MockHidapiDevice):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._fan_configs = (4, 20, 40, 50, 60, 70, 80, 90)
        self._fan_temp_configs = (4, 30, 40, 50, 60, 70, 80, 90)
        self._model_idx = 255  # TODO: check the correct model index from device
        self._feature_data = Report(_DEFAULT_FEATURE_DATA[0], _DEFAULT_FEATURE_DATA[1:])

        self.preload_read(self._feature_data)

    def write(self, data):
        reply = bytearray(_REPORT_LENGTH)
        reply[0:2] = data[0:2]
        if list(data[:2]) == [0x01, 0xB1]:  # get current model idx request
            reply[2] = self._model_idx
        elif list(data[:2]) == [0xD0, 0x31]:  # get status request
            reply[2:23] = (
                pack("<h", 496)
                + pack("<h", 517)
                + pack("<h", 509)
                + pack("<h", 1045)
                + pack("<h", 1754)
                + bytearray([0, 0, 0, 0, 0x7D, 0, 0x7D, 0, 0, 0])
                + pack("<h", 20)
                + pack("<h", 20)
                + pack("<h", 20)
                + pack("<h", 20)
                + pack("<h", 50)
            )
            self.preload_read(Report(0, reply))
        elif list(data[:2]) == [0xD0, 0x32]:  # get fan config request
            for i in (2, 10, 18, 26, 34):
                reply[i : i + 10] = self._fan_configs
            self.preload_read(Report(0, reply))
        elif list(data[:2]) == [0xD0, 0x33]:  # get fan T config request
            reply[1] = 0x32
            for i in (2, 10, 18, 26, 34):
                reply[i : i + 10] = self._fan_temp_configs
            self.preload_read(Report(0, reply))
        elif list(data[1:3]) == [0xB0, 0xCC]:  # get ldrom fw
            self.preload_read(Report(0, reply))
        elif list(data[1:3]) == [0xB6, 0xCC]:  # get aprom fw
            self.preload_read(Report(0, reply))
        elif data[1] == 0xF1:  # get screen fw
            self.preload_read(Report(0, reply))
        return super().write(data)


class _MockCoreLiquidInvalid(MockHidapiDevice):
    def read(self, length):
        buf = bytearray([0xD0, 0x32] + (62 * [0]))
        return buf[:length]


def test_mpg_core_liquid_k360_initializes(mpgCoreLiquidK360Device):
    mpgCoreLiquidK360Device.initialize()

    writes = len(mpgCoreLiquidK360Device.device.sent)
    assert (
        writes
        == 9
        # 1 get fan config,
        # 2 get fan T config,
        # 3 get aprom fw,
        # 4 get ldrom fw,
        # 5 get display fw,
        # 6 set screen settings,
        # 7 set fan config,
        # 8 set fan T config,
        # 9 set safe control temperature
    )


def test_mpg_core_liquid_k360_get_status(mpgCoreLiquidK360Device):
    dev = mpgCoreLiquidK360Device
    (
        fan1,
        fan1d,
        fan2,
        fan2d,
        fan3,
        fan3d,
        wbfan,
        wbfand,
        pump,
        pumpd,
    ) = dev.get_status()
    assert fan1[1] == 496
    assert fan1d[1] == 20
    assert fan2[1] == 517
    assert fan2d[1] == 20
    assert fan3[1] == 509
    assert fan3d[1] == 20
    assert wbfan[1] == 1045
    assert wbfand[1] == 20
    assert pump[1] == 1754
    assert pumpd[1] == 50
    assert dev.device.sent[-1].number == 0xD0
    assert dev.device.sent[-1].data[0] == 0x31


def test_mpg_core_liquid_k360_get_status_invalid_read(mpgCoreLiquidK360DeviceInvalid):
    dev = mpgCoreLiquidK360DeviceInvalid
    with pytest.raises(AssertionError):
        dev.get_status()


def test_mpg_core_liquid_k360_set_fixed_speeds(mpgCoreLiquidK360Device):
    mpgCoreLiquidK360Device.set_fixed_speed("pump", 65)

    fan_report, T_report = mpgCoreLiquidK360Device.device.sent[-2:]
    assert fan_report.data[33:41] == [3] + [65] * 7


def test_mpg_core_liquid_k360_set_speed_profile(mpgCoreLiquidK360Device):
    duties = [20, 30, 34, 40, 50]
    temps = [30, 50, 80, 90, 100]
    curve_profile = zip(duties, temps)

    mpgCoreLiquidK360Device.set_speed_profile("fans", curve_profile)

    fan_report, T_report = mpgCoreLiquidK360Device.device.sent[-3:-1]

    # fan 1
    assert fan_report.data[1:9] == [3] + duties + [0] * 2
    assert T_report.data[1:9] == [3] + temps + [0] * 2
    # fan 2
    assert fan_report.data[9:17] == [3] + duties + [0] * 2
    assert T_report.data[9:17] == [3] + temps + [0] * 2
    # fan 3
    assert fan_report.data[17:25] == [3] + duties + [0] * 2
    assert T_report.data[17:25] == [3] + temps + [0] * 2


def test_mpg_core_liquid_k360_set_color(mpgCoreLiquidK360Device):
    colors = [[255, 255, 0], [0, 255, 255]]
    mode = "clock"
    speed = 2
    brightness = 5
    use_color = 1

    mpgCoreLiquidK360Device.set_color(
        "sync", mode, colors, speed=speed, brightness=brightness, color_selection=use_color
    )
    report = mpgCoreLiquidK360Device.device.sent[-1]

    assert report.data[30] == 20  # mode
    assert report.data[34] == (brightness << 2) | speed  # brightness, speed
    assert report.data[31:34] == colors[0]
    assert report.data[35:38] == colors[1]


def test_mpg_core_liquid_k360_not_totally_broken(mpgCoreLiquidK360Device):
    """Reasonable calls to untested APIs do not raise exceptions"""
    dev = mpgCoreLiquidK360Device
    dev.initialize()
    _ = dev.get_status()

    profile = ((0, 30), (25, 40), (60, 60), (100, 75))
    dev.set_speed_profile("pump", profile)
    dev.set_fixed_speed("waterblock-fan", 42)
    dev.set_screen("lcd", "image", "0;4")
    dev.set_screen("lcd", "banner", "1;0;Hello, world")
    dev.set_screen("lcd", "disable", "")
    dev.set_screen("lcd", "clock", "0")
    dev.set_screen("lcd", "hardware", "cpu_temp;cpu_freq")


def test_mpg_core_liquid_k360_set_clock(mpgCoreLiquidK360Device):
    time = datetime(2012, 12, 21, 9, 54, 20)

    mpgCoreLiquidK360Device.set_time(time)

    report = mpgCoreLiquidK360Device.device.sent[-1]
    assert report.data[:8] == [131, 12, 12, 21, 4, 9, 54, 20]


def test_mpg_core_liquid_k360_set_hw_status(mpgCoreLiquidK360Device):
    cpu_freq = 3500.0
    cpu_T = 54.0
    gpu_f = 7000

    mpgCoreLiquidK360Device.set_hardware_status(cpu_T, cpu_f=cpu_freq, gpu_f=gpu_f)

    cpu_report = mpgCoreLiquidK360Device.device.sent[-2]
    assert cpu_report.data[:5] == [133, 172, 13, 54, 0]

    gpu_report = mpgCoreLiquidK360Device.device.sent[-1]
    assert gpu_report.data[:5] == [134, 88, 27, 0, 0]


def test_unsafe_core_liquid_get_status(mpgCoreLiquidDeviceExperimental):
    status = mpgCoreLiquidDeviceExperimental.get_status()
    assert status == []

    status = mpgCoreLiquidDeviceExperimental.get_status(unsafe=["other"])
    assert status == []

    status = mpgCoreLiquidDeviceExperimental.get_status(unsafe=["experimental_coreliquid_cooler"])
    assert mpgCoreLiquidDeviceExperimental.device.sent[-1].number == 0xD0


def test_unsafe_core_liquid_set_fixed_speed(mpgCoreLiquidDeviceExperimental):
    with pytest.raises(UnsafeFeaturesNotEnabled):
        mpgCoreLiquidDeviceExperimental.set_fixed_speed("pump", 65)

    with pytest.raises(UnsafeFeaturesNotEnabled):
        mpgCoreLiquidDeviceExperimental.set_fixed_speed("pump", 65, unsafe=["other"])

    mpgCoreLiquidDeviceExperimental.set_fixed_speed(
        "pump", 65, unsafe=["experimental_coreliquid_cooler"]
    )


def test_unsafe_core_liquid_set_speed_profile(mpgCoreLiquidDeviceExperimental):
    duties = [20, 30, 34, 40, 50]
    temps = [30, 50, 80, 90, 100]
    curve_profile = zip(duties, temps)
    with pytest.raises(UnsafeFeaturesNotEnabled):
        mpgCoreLiquidDeviceExperimental.set_speed_profile("fans", curve_profile)

    with pytest.raises(UnsafeFeaturesNotEnabled):
        mpgCoreLiquidDeviceExperimental.set_speed_profile("fans", curve_profile, unsafe=["other"])

    mpgCoreLiquidDeviceExperimental.set_speed_profile(
        "fans", curve_profile, unsafe=["experimental_coreliquid_cooler"]
    )


def test_unsafe_core_liquid_set_color(mpgCoreLiquidDeviceExperimental):
    colors = [[255, 255, 0], [0, 255, 255]]
    mode = "clock"
    speed = 2
    brightness = 5
    use_color = 1

    with pytest.raises(UnsafeFeaturesNotEnabled):
        mpgCoreLiquidDeviceExperimental.set_color(
            "sync", mode, colors, speed=speed, brightness=brightness, color_selection=use_color
        )

    with pytest.raises(UnsafeFeaturesNotEnabled):
        mpgCoreLiquidDeviceExperimental.set_color(
            "sync",
            mode,
            colors,
            speed=speed,
            brightness=brightness,
            color_selection=use_color,
            unsafe=["other"],
        )

    mpgCoreLiquidDeviceExperimental.set_color(
        "sync",
        mode,
        colors,
        speed=speed,
        brightness=brightness,
        color_selection=use_color,
        unsafe=["experimental_coreliquid_cooler"],
    )


def test_unsafe_core_liquid_set_clock(mpgCoreLiquidDeviceExperimental):
    time = datetime(2012, 12, 21, 9, 54, 20)

    with pytest.raises(UnsafeFeaturesNotEnabled):
        mpgCoreLiquidDeviceExperimental.set_time(time)

    with pytest.raises(UnsafeFeaturesNotEnabled):
        mpgCoreLiquidDeviceExperimental.set_time(time, unsafe=["other"])

    mpgCoreLiquidDeviceExperimental.set_time(time, unsafe=["experimental_coreliquid_cooler"])


def test_unsafe_core_liquid_set_hw_status(mpgCoreLiquidDeviceExperimental):
    cpu_freq = 3500.0
    cpu_T = 54.0

    with pytest.raises(UnsafeFeaturesNotEnabled):
        mpgCoreLiquidDeviceExperimental.set_hardware_status(cpu_T, cpu_f=cpu_freq)

    with pytest.raises(UnsafeFeaturesNotEnabled):
        mpgCoreLiquidDeviceExperimental.set_hardware_status(cpu_T, cpu_f=cpu_freq, unsafe=["other"])

    mpgCoreLiquidDeviceExperimental.set_hardware_status(
        cpu_T, cpu_f=cpu_freq, unsafe=["experimental_coreliquid_cooler"]
    )



#!/usr/bin/env python3

"""prometheus-liquidctl-exporter – host a metrics HTTP endpoint with Prometheus formatted data from liquidctl

This is an experimental script that collects stats from liquidctl and exposes them as a http://localhost:8098/metrics
endpoint in the Prometheus text format.
See: https://prometheus.io/docs/instrumenting/exposition_formats/#text-format-example

Example metric with labels:
# HELP liquidctl liquidctl exported metrics
# TYPE liquidctl gauge
liquidctl{device="NZXT Kraken X (X42, X52, X62 or X72)",sensor="liquid_temperature",unit="°C"} 33.6

Usage:
  prometheus-liquidctl-exporter [options]

Options:
  --legacy-690lc                     Use Asetek 690LC in legacy mode (old Krakens)
  --server-port <number>             Port for the HTTP /metrics endpoint
  -v, --verbose                      Output additional information
  -g, --debug                        Show debug information on stderr
  --version                          Display the version number
  --help                             Show this message

Device selection options (see: list -v):
  -m, --match <substring>            Filter devices by description substring
  -n, --pick <number>                Pick among many results for a given filter
  --vendor <id>                      Filter devices by hexadecimal vendor ID
  --product <id>                     Filter devices by hexadecimal product ID
  --release <number>                 Filter devices by hexadecimal release number
  --serial <number>                  Filter devices by serial number
  --bus <bus>                        Filter devices by bus
  --address <address>                Filter devices by address in bus
  --usb-port <port>                  Filter devices by USB port in bus


Copyright Alex Berryman, Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
import sys
import time
import usb
from datetime import timedelta

from docopt import docopt
from liquidctl.driver import *
from prometheus_client import start_http_server
from prometheus_client.core import GaugeMetricFamily, REGISTRY, InfoMetricFamily

LOGGER = logging.getLogger(__name__)


def gauge_name_sanitize(name):
    return name.replace(" ", "_").lower()


class LiquidCollector(object):
    def __init__(self):
        self.description = "liquidctl exported metrics"

    def collect(self):
        labels = ["device", "sensor", "unit"]
        g = GaugeMetricFamily("liquidctl", self.description, labels=labels)
        i = InfoMetricFamily("liquidctl", self.description, labels=["device"])
        for d in devs:
            try:
                get_status = d.get_status()
                for metric in get_status:
                    sanitized_name = gauge_name_sanitize(metric[0])
                    sample_value = metric[1]
                    unit = metric[2]

                    if isinstance(sample_value, timedelta):
                        # cast timedelta into seconds and override the supplied unit
                        sample_value = sample_value.seconds
                        unit = "seconds"

                    if unit != "":
                        # FIXME doesn't handle multiple equal devices well
                        label_values = [d.description, sanitized_name, unit]
                        g.add_metric(label_values, value=sample_value)
                        LOGGER.debug(
                            "%s: %s as GaugeMetric %s labels %s",
                            d.description,
                            metric,
                            sanitized_name,
                            "/".join(label_values),
                        )
                    else:
                        i.add_metric([d.description], value={sanitized_name: sample_value})
                        LOGGER.debug(
                            "%s: %s InfoMetric labeled with %s => %s",
                            d.description,
                            metric,
                            sanitized_name,
                            sample_value,
                        )
            except usb.core.USBError as err:
                LOGGER.warning("failed to read from the device, possibly serving stale data")
                LOGGER.debug(err, exc_info=True)
        yield g
        yield i


def _make_opts(arguments):
    options = {}
    for arg, val in arguments.items():
        if val is not None and arg in _PARSE_ARG:
            opt = arg.replace("--", "").replace("-", "_")
            options[opt] = _PARSE_ARG[arg](val)
    return options


_PARSE_ARG = {
    "--legacy-690lc": bool,
    "--vendor": lambda x: int(x, 16),
    "--product": lambda x: int(x, 16),
    "--release": lambda x: int(x, 16),
    "--serial": str,
    "--bus": str,
    "--address": str,
    "--usb-port": lambda x: tuple(map(int, x.split("."))),
    "--match": str,
    "--pick": int,
}

if __name__ == "__main__":
    args = docopt(__doc__, version="0.1.1")
    opts = _make_opts(args)
    devs = list(find_liquidctl_devices(**opts))
    for d in devs:
        LOGGER.info("initializing %s", d.description)
        d.connect()

    if args["--debug"]:
        args["--verbose"] = True
        logging.basicConfig(level=logging.DEBUG, format="[%(levelname)s] %(name)s: %(message)s")
    elif args["--verbose"]:
        logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
    else:
        logging.basicConfig(level=logging.WARNING, format="%(message)s")
        sys.tracebacklimit = 0

    REGISTRY.register(LiquidCollector())

    if args["--server-port"]:
        server_port = int(args["--server-port"])
    else:
        server_port = 8098

    start_http_server(server_port)
    LOGGER.debug("server started on port %s", server_port)

    try:
        while True:
            # Keep HTTP server alive in a loop
            time.sleep(2)
    except KeyboardInterrupt:
        LOGGER.info("canceled by user")
    finally:
        for d in devs:
            try:
                LOGGER.info("disconnecting from %s", d.description)
                d.disconnect()
            except:
                LOGGER.exception("unexpected error when disconnecting")



import multiprocessing
import os
import pytest
import sys
import time
from pathlib import Path

from liquidctl.keyval import RuntimeStorage, _FilesystemBackend

mp_ctx = multiprocessing.get_context("spawn")


@pytest.fixture
def tmpstore(tmpdir):
    run_dir = tmpdir.mkdir("run_dir")
    prefixes = ["prefix"]

    backend = _FilesystemBackend(key_prefixes=prefixes, runtime_dirs=[run_dir])
    return RuntimeStorage(prefixes, backend=backend)


def test_loads_and_stores(tmpstore):
    assert tmpstore.load("key") is None
    assert tmpstore.load("key", default=42) == 42

    tmpstore.store("key", "42")

    assert tmpstore.load("key") == "42"
    assert tmpstore.load("key", of_type=int) is None


def test_updates_with_load_store(tmpstore):
    assert tmpstore.load_store("key", lambda x: x) == (None, None)
    assert tmpstore.load_store("key", lambda x: x, default=42) == (None, 42)
    assert tmpstore.load_store("key", lambda x: str(x)) == (42, "42")
    assert tmpstore.load_store("key", lambda x: x, of_type=int) == ("42", None)


def test_fs_backend_stores_truncate_appropriately(tmpdir):
    run_dir = tmpdir.mkdir("run_dir")

    # use a separate reader to prevent caching from masking issues
    writer = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir])
    reader = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir])

    writer.store("key", 42)
    assert reader.load("key") == 42

    writer.store("key", 1)
    assert reader.load("key") == 1

    writer.load_store("key", lambda _: 42)
    assert reader.load("key") == 42

    writer.load_store("key", lambda _: 1)
    assert reader.load("key") == 1


def test_fs_backend_loads_from_fallback_dir(tmpdir):
    run_dir = tmpdir.mkdir("run_dir")
    fb_dir = tmpdir.mkdir("fb_dir")

    fallback = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[fb_dir])
    fallback.store("key", 42)

    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir, fb_dir])
    assert store.load("key") == 42

    store.store("key", -1)
    assert store.load("key") == -1
    assert fallback.load("key") == 42, "fallback location was changed"


def test_fs_backend_handles_values_corupted_with_nulls(tmpdir, caplog):
    run_dir = tmpdir.mkdir("run_dir")
    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir])

    store.store("key", 42)
    key_file = Path(run_dir).joinpath("prefix", "key")
    assert key_file.read_bytes() == b"42", "unit test is unsound"

    key_file.write_bytes(b"\x00")
    val = store.load("key")

    assert val is None
    assert "was corrupted" in caplog.text

    val, new_val = store.load_store("key", lambda x: 24)
    assert val is None
    assert new_val == 24


def test_fs_backend_load_store_returns_old_and_new_values(tmpdir):
    run_dir = tmpdir.mkdir("run_dir")

    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir])

    assert store.load_store("key", lambda _: 42) == (None, 42)
    assert store.load_store("key", lambda x: x + 1) == (42, 43)


def test_fs_backend_load_store_loads_from_fallback_dir(tmpdir):
    run_dir = tmpdir.mkdir("run_dir")
    fb_dir = tmpdir.mkdir("fb_dir")

    fallback = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[fb_dir])
    fallback.store("key", 42)

    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir, fb_dir])
    assert store.load_store("key", lambda x: x + 1) == (42, 43)

    assert fallback.load("key") == 42, "content in fallback location changed"


def test_fs_backend_load_store_loads_from_fallback_dir_that_is_symlink(tmpdir):
    # should deadlock if there is a problem with the lock type or with the
    # handling of fallback paths that point to the same principal/write
    # directory

    run_dir = tmpdir.mkdir("run_dir")
    fb_dir = os.path.join(run_dir, "symlink")
    try:
        os.symlink(run_dir, fb_dir, target_is_directory=True)
    except OSError as _:
        if sys.platform == "win32":
            pytest.skip("unable to create Windows symlink with current permissions")
        else:
            raise

    # don't store any initial value so that the fallback location is checked

    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir, fb_dir])
    assert store.load_store("key", lambda x: 42) == (None, 42)

    fallback = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[fb_dir])
    assert fallback.load("key") == 42, "content in fallback symlink did not change"


def test_fs_backend_load_store_is_atomic(tmpdir):
    run_dir = tmpdir.mkdir("run_dir")

    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir])
    store.store("key", 42)

    ps = [
        mp_ctx.Process(target=_fs_mp_increment_key, args=(run_dir, "prefix", "key", 0.2)),
        mp_ctx.Process(target=_fs_mp_increment_key, args=(run_dir, "prefix", "key", 0.2)),
        mp_ctx.Process(target=_fs_mp_increment_key, args=(run_dir, "prefix", "key", 0.2)),
    ]

    start_time = time.monotonic()

    for p in ps:
        p.start()

    for p in ps:
        p.join()

    elapsed = time.monotonic() - start_time

    assert store.load("key") == 45
    assert elapsed >= 0.2 * len(ps)


def test_fs_backend_loads_honor_load_store_locking(tmpdir):
    run_dir = tmpdir.mkdir("run_dir")

    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir])
    store.store("key", 42)

    ps = [
        mp_ctx.Process(target=_fs_mp_increment_key, args=(run_dir, "prefix", "key", 0.2)),
        mp_ctx.Process(target=_fs_mp_check_key, args=(run_dir, "prefix", "key", 43)),
    ]

    ps[0].start()
    time.sleep(0.1)
    ps[1].start()

    for p in ps:
        p.join()


def test_fs_backend_stores_honor_load_store_locking(tmpdir):
    run_dir = tmpdir.mkdir("run_dir")

    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir])
    store.store("key", 42)

    ps = [
        mp_ctx.Process(target=_fs_mp_increment_key, args=(run_dir, "prefix", "key", 0.2)),
        mp_ctx.Process(target=_fs_mp_store_key, args=(run_dir, "prefix", "key", -1)),
    ]

    start_time = time.monotonic()

    ps[0].start()
    time.sleep(0.1)
    ps[1].start()

    # join second process first
    ps[1].join()

    elapsed = time.monotonic() - start_time
    assert elapsed >= 0.2

    ps[0].join()
    assert store.load("key") == -1


def test_fs_backend_releases_locks(tmpdir):
    # should deadlock if any method does not properly release its lock

    run_dir = tmpdir.mkdir("run_dir")
    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir])

    def incr_from_other_process():
        other = mp_ctx.Process(target=_fs_mp_increment_key, args=(run_dir, "prefix", "key", 0.0))
        other.start()
        other.join()

    store.store("key", 42)
    incr_from_other_process()
    assert store.load("key") == 43

    store.load_store("key", lambda _: -1)
    incr_from_other_process()
    assert store.load("key") == 0

    incr_from_other_process()
    assert store.load("key") == 1


def _fs_mp_increment_key(run_dir, prefix, key, sleep):
    """Open a _FilesystemBackend and increment `key`.

    For the `multiprocessing` tests.

    Opens the storage on `run_dir` and with `prefix`.  Sleeps for `sleep`
    seconds within the increment closure.
    """

    def l(x):
        time.sleep(sleep)
        return x + 1

    store = _FilesystemBackend(key_prefixes=[prefix], runtime_dirs=[run_dir])
    store.load_store(key, l)


def _fs_mp_check_key(run_dir, prefix, key, expected):
    """Open a _FilesystemBackend and check `key` value against `expected`.

    For the `multiprocessing` tests.

    Opens the storage on `run_dir` and with `prefix`.
    """

    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir])
    assert store.load(key) == expected


def _fs_mp_store_key(run_dir, prefix, key, new_value):
    """Open a _FilesystemBackend and store `new_value` for `key`.

    For the `multiprocessing` tests.

    Opens the storage on `run_dir` and with `prefix`.
    """

    store = _FilesystemBackend(key_prefixes=["prefix"], runtime_dirs=[run_dir])
    store.store(key, new_value)



"""Test backward compatibility with liquidctl 1.12.x."""

# uses the psf/black style

import pytest
from _testutils import MockHidapiDevice, Report, MockRuntimeStorage

from test_commander_pro import commanderProDevice


def test_initialize_commander_pro_fan_modes(commanderProDevice, caplog):
    """Fix #615 but preserve API compatibility."""

    responses = [
        "000009d4000000000000000000000000",  # firmware
        "00000500000000000000000000000000",  # bootloader
        "00010100010000000000000000000000",  # temp probes
        "00010102000000000000000000000000",  # fan set (throw away)
        "00010102000000000000000000000000",  # fan set (throw away)
        "00010102000000000000000000000000",  # fan set (throw away)
        "00010102000000000000000000000000",  # fan probes
    ]
    for d in responses:
        commanderProDevice.device.preload_read(Report(0, bytes.fromhex(d)))

    commanderProDevice.initialize(direct_access=True, fan_modes={"4": "dc"})

    sent = commanderProDevice.device.sent
    assert len(sent) == 5
    assert sent[3].data[0] == 0x28
    assert sent[3].data[2] == 3
    assert sent[3].data[3] == 1

    assert "deprecated parameter name `fan_modes`" in caplog.text



import sys

collect_ignore = [
    "setup.py",
    "extra/contrib/fusion_rgb_cycle.py",  # depends on coloraide
    "extra/prometheus-liquidctl-exporter.py",  # depends on prometheus_client
]

if sys.platform != "linux":
    collect_ignore.append("tests/test_smbus.py")

if sys.platform not in ["win32", "cygwin"]:
    collect_ignore.append("extra/windows/LQiNFO.py")



#!/usr/bin/env python3

"""yoda – dynamically adjust liquidctl device pump and fan speeds.

Periodically adjusts pump and fan speeds according to user-specified profiles.

Different sensors can be used for each channel.  Use show-sensors to view the
sensors available for use with a particular device.

To avoid jerks in pump or fan speeds, an exponential moving average is used as
low-pass filter on sensor data.

Profiles are specified as comma-separated lists of `(temperature,duty)` pairs.
For example: `(20,50),(40,65),(40,65),(50,100)` specifies a duty of 65% at
40°C.  The profile will be linearly interpolated between the specified points.

In device controlled mode, sets the device internal control profile and periodically
sends sensor data, but the device will independently control duty cycles to match the temperature.
Named profiles implemented by the device manufacturer, such as "silent", "game", "smart",
are only available in device controlled mode.

Escape sequences or appropriate single or double quotes should be employed to
escape characters that are reserved by the shell in use (e.g. in the case of
bash, the parenthesis and any optional whitespace).  In practice, wrapping the
profile in double quotes should be sufficient for most users.

Examples:
  yoda --match grid show-sensors
  yoda --match grid control fan1 with "(20,20),(35,100)" on nct6793.systin
  yoda --match kraken show-sensors
  yoda --match kraken control pump with "(20,50),(50,100)" on istats.cpu and fan with "(20,25),(34,100)" on _internal.liquid
  yoda --match msi control pump with "smart" on coretemp.package_id_0 and fans with "silent" on coretemp.package_id_0

Usage:
  yoda [options] show-sensors
  yoda [options] control (<channel> with <profile> on <sensor> [and])...
  yoda --help
  yoda --version

Options:
  --interval <seconds>     Update interval in seconds [default: 2]
  -m, --match <substring>  Filter devices by description substring
  -n, --pick <number>      Pick among many results for a given filter
  --vendor <id>            Filter devices by vendor id
  --product <id>           Filter devices by product id
  --release <number>       Filter devices by release number
  --serial <number>        Filter devices by serial number
  --bus <bus>              Filter devices by bus
  --address <address>      Filter devices by address in bus
  --usb-port <port>        Filter devices by USB port in bus
  --unsafe <features>      Comma-separated bleeding-edge features to enable
  -v, --verbose            Output additional information
  -g, --debug              Show debug information on stderr
  --legacy-690lc           Use Asetek 690LC in legacy mode (old Krakens)
  --use-device-controller  Use the control loop integrated to the device (MPG coreliquid device)
  --version                Display the version number
  --help                   Show this message

Requirements:
  all platforms  liquidctl, including the Python APIs (pip install liquidctl)
  Linux/FreeBSD  psutil [optional] (pip install psutil)
  macOS          iStats (gem install iStats)
  Windows        none, system sensors not yet supported

Changelog:
  0.0.6  Add --use-device-controller and make psutil optional.
  0.0.5  Document how profiles are specified
  0.0.4  Fix casing of log and error messages
  0.0.3  Remove duplicate option definition
  0.0.2  Add low-pass filter and basic error handling.
  0.0.1  Generalization of krakencurve-poc 0.0.2 to multiple devices.

Copyright Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""


import ast
import logging
import math
import sys
import time
from datetime import datetime

from docopt import docopt
import liquidctl.cli as _borrow
from liquidctl.util import normalize_profile, interpolate_profile
import liquidctl.driver

VERSION = "0.0.6"

LOGGER = logging.getLogger(__name__)

INTERNAL_CHIP_NAME = "_internal"

MAX_FAILURES = 3

if sys.platform == "darwin":
    import re
    import subprocess
elif sys.platform.startswith("linux") or sys.platform.startswith("freebsd"):
    try:
        import psutil
    except ModuleNotFoundError:
        psutil = None


def read_sensors(device, **kwargs):
    sensors = {}
    for k, v, u in device.get_status(**kwargs):
        if u == "°C":
            sensor_name = k.lower().replace(" ", "_").replace("_temperature", "")
            sensors[f"{INTERNAL_CHIP_NAME}.{sensor_name}"] = v
    if sys.platform == "darwin":
        istats_stdout = subprocess.check_output(["istats"]).decode("utf-8")
        for line in istats_stdout.split("\n"):
            if line.startswith("CPU"):
                cpu_temp = float(re.search(r"\d+\.\d+", line).group(0))
                sensors["istats.cpu"] = cpu_temp
                break
    elif psutil:
        for m, li in psutil.sensors_temperatures().items():
            for label, current, _, _ in li:
                sensor_name = label.lower().replace(" ", "_")
                sensors[f"{m}.{sensor_name}"] = current
        sensors["cpu_freq"] = psutil.cpu_freq().current
    return sensors


def show_sensors(device, **kwargs):
    print("{:<60}  {:>14}".format("Sensor identifier", "Value"))
    print("-" * 80)
    sensors = read_sensors(device, **kwargs)
    for k, v in sensors.items():
        unit = "MHz" if k == "cpu_freq" else "°C"
        print(f"{k:<60}  {v:>14.1f} {unit}")


def parse_profile(arg, mintemp=0, maxtemp=100, minduty=0, maxduty=100, str_allowed=False):
    """Parse, validate and normalize a temperature–duty profile.

    >>> parse_profile('smart', str_allowed=True)
    'smart'

    >>> parse_profile('(20,30),(30,50),(34,80),(40,90)', 0, 60, 25, 100)
    [(20, 30), (30, 50), (34, 80), (40, 90), (60, 100)]
    >>> parse_profile('35', 0, 60, 25, 100)
    [(0, 35), (59, 35), (60, 100)]

    The profile is validated in structure and acceptable ranges.  Duty is
    checked against `minduty` and `maxduty`.  Temperature must be between
    `mintemp` and `maxtemp`.

    >>> parse_profile('(20,30),(50,100', 0, 60, 25, 100)
    Traceback (most recent call last):
        ...
    ValueError: profile must be comma-separated (temperature, duty) tuples or supported mode name
    >>> parse_profile('(20,30),(50,100,2)', 0, 60, 25, 100)
    Traceback (most recent call last):
        ...
    ValueError: profile must be comma-separated (temperature, duty) tuples
    >>> parse_profile('(20,30),(50,97.6)', 0, 60, 25, 100)
    Traceback (most recent call last):
        ...
    ValueError: duty must be integer between 25 and 100
    >>> parse_profile('(20,15),(50,100)', 0, 60, 25, 100)
    Traceback (most recent call last):
        ...
    ValueError: duty must be integer between 25 and 100
    >>> parse_profile('(20,30),(70,100)', 0, 60, 25, 100)
    Traceback (most recent call last):
        ...
    ValueError: temperature must be integer between 0 and 60

    """
    try:
        if str_allowed and arg in liquidctl.driver.msi.MpgCooler.BUILTIN_MODES:
            return arg
        else:
            val = ast.literal_eval("[" + arg + "]")
            if len(val) == 1 and isinstance(val[0], int):
                # for arg == '<number>' set fixed duty between mintemp and maxtemp - 1
                val = [(mintemp, val[0]), (maxtemp - 1, val[0])]
    except:
        raise ValueError(
            "profile must be comma-separated (temperature, duty) tuples or supported mode name"
        )
    for step in val:
        if not isinstance(step, tuple) or len(step) != 2:
            raise ValueError("profile must be comma-separated (temperature, duty) tuples")
        temp, duty = step
        if not isinstance(temp, int) or temp < mintemp or temp > maxtemp:
            raise ValueError(
                "temperature must be integer between {} and {}".format(mintemp, maxtemp)
            )
        if not isinstance(duty, int) or duty < minduty or duty > maxduty:
            raise ValueError("duty must be integer between {} and {}".format(minduty, maxduty))
    return normalize_profile(val, critx=maxtemp)


def auto_control(device, channels, profiles, sensors, update_interval, **kwargs):
    """Communicate sensor data directly with the device.

    Implemented for use with the MSI coreliquid AIO.
    Allows compatible devices to utilize their internal control loop
    to determine appropriate fan speeds for the CPU temperature.
    """
    assert getattr(
        device, "HAS_AUTOCONTROL", False
    ), f"No registered control loop capability for device {device}!"

    device.set_profiles(channels, profiles, **kwargs)

    assert all(
        s == sensors[0] for s in sensors
    ), "Controlling different channels with different sensors not possible with device control"
    sensor = sensors[0]

    LOGGER.info("starting...")
    failures = 0
    while True:
        try:
            sensor_data = read_sensors(device, **kwargs)
            temp = sensor_data[sensor]
            freq = sensor_data.get("cpu_freq", 0)

            device.set_time(datetime.now(), **kwargs)
            device.set_hardware_status(
                temp,
                cpu_f=freq,
                gpu_f=sensor_data.get("gpu_freq", 0),
                gpu_U=sensor_data.get("gpu_usage", 0),
                **kwargs,
            )
            failures = 0
        except Exception as err:
            failures += 1
            LOGGER.error(err)
            if failures >= MAX_FAILURES:
                LOGGER.critical("too many failures in a row: %d", failures)
                raise
        time.sleep(update_interval)


def control(device, channels, profiles, sensors, update_interval, **kwargs):
    LOGGER.info(
        "device: %s on bus %s and address %s", device.description, device.bus, device.address
    )
    for channel, profile, sensor in zip(channels, profiles, sensors):
        LOGGER.info("channel: %s following profile %s on %s", channel, str(profile), sensor)

    averages = [None] * len(channels)
    cutoff_freq = 1 / update_interval / 10
    alpha = 1 - math.exp(-2 * math.pi * cutoff_freq)
    LOGGER.info(
        "update interval: %d s; cutoff frequency (low-pass): %.2f Hz; ema alpha: %.2f",
        update_interval,
        cutoff_freq,
        alpha,
    )

    try:
        # more efficient and safer API, but only supported by very few devices
        apply_duty = device.set_instantaneous_speed
    except AttributeError:
        apply_duty = device.set_fixed_speed

    LOGGER.info("starting...")
    failures = 0
    while True:
        try:
            sensor_data = read_sensors(device, **kwargs)
            for i, (channel, profile, sensor) in enumerate(zip(channels, profiles, sensors)):
                # compute the exponential moving average (ema), used as a low-pass filter (lpf)
                ema = averages[i]
                sample = sensor_data[sensor]
                if ema is None:
                    ema = sample
                else:
                    ema = alpha * sample + (1 - alpha) * ema
                averages[i] = ema

                # interpolate on sensor ema and apply corresponding duty
                duty = interpolate_profile(profile, ema)
                LOGGER.info(
                    "%s control: lpf(%s) = lpf(%.1f°C) = %.1f°C => duty := %d%%",
                    channel,
                    sensor,
                    sample,
                    ema,
                    duty,
                )
                apply_duty(channel, duty, **kwargs)
            if getattr(device, "NEEDS_TIME", False):
                device.set_time(datetime.now(), **kwargs)
            if getattr(device, "NEEDS_HWSTATUS", False):
                device.set_hardware_status(
                    sensor_data[sensors[0]],
                    cpu_f=sensor_data.get("cpu_freq", 0),
                    gpu_f=sensor_data.get("gpu_freq", 0),
                    gpu_U=sensor_data.get("gpu_usage", 0),
                    **kwargs,
                )

            failures = 0
        except Exception as err:
            failures += 1
            LOGGER.error(err)
            if failures >= MAX_FAILURES:
                LOGGER.critical("too many failures in a row: %d", failures)
                raise
        time.sleep(update_interval)


if __name__ == "__main__":
    if len(sys.argv) == 2 and sys.argv[1] == "doctest":
        import doctest

        doctest.testmod(verbose=True)
        sys.exit(0)

    args = docopt(__doc__, version="yoda v{}".format(VERSION))

    if args["--debug"]:
        args["--verbose"] = True
        logging.basicConfig(level=logging.DEBUG, format="[%(levelname)s] %(name)s: %(message)s")
        import liquidctl.version

        LOGGER.debug("yoda v%s", VERSION)
        LOGGER.debug("liquidctl v%s", liquidctl.version.__version__)
    elif args["--verbose"]:
        logging.basicConfig(level=logging.WARNING, format="%(levelname)s: %(message)s")
        LOGGER.setLevel(logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING, format="%(levelname)s: %(message)s")
        sys.tracebacklimit = 0
    if args["--unsafe"] is not None:
        unsafe = args["--unsafe"].lower().split(",")
    else:
        unsafe = []

    if (sys.platform.startswith("linux") or sys.platform.startswith("freebsd")) and not psutil:
        LOGGER.warning("system sensors are not available, psutil not found")

    frwd = _borrow._make_opts(args)
    selected = list(liquidctl.driver.find_liquidctl_devices(**frwd))
    if len(selected) > 1:
        raise SystemExit(
            "too many devices, filter or select one.  See liquidctl --help and yoda --help."
        )
    elif len(selected) == 0:
        raise SystemExit("no devices matches available drivers and selection criteria")

    device = selected[0]
    device.connect(unsafe=unsafe)
    try:
        if args["show-sensors"]:
            show_sensors(device, unsafe=unsafe)
        elif args["control"]:
            if args["--use-device-controller"]:
                auto_control(
                    device,
                    args["<channel>"],
                    list(map(lambda p: parse_profile(p, str_allowed=True), args["<profile>"])),
                    args["<sensor>"],
                    update_interval=int(args["--interval"]),
                    unsafe=unsafe,
                )
            else:
                control(
                    device,
                    args["<channel>"],
                    list(map(parse_profile, args["<profile>"])),
                    args["<sensor>"],
                    update_interval=int(args["--interval"]),
                    unsafe=unsafe,
                )
        else:
            raise Exception("nothing to do")
    except KeyboardInterrupt:
        LOGGER.info("stopped by user.")
    finally:
        device.disconnect()



#!/usr/bin/env python3

"""liquiddump – continuously dump monitoring data from liquidctl devices.

This is a experimental script that continuously dumps the status of all
available devices to stdout in newline-delimited JSON.

Usage:
  liquiddump [options]
  liquiddump --help
  liquiddump --version

Options:
  --interval <seconds>    Update interval in seconds [default: 2]
  --legacy-690lc          Use Asetek 690LC in legacy mode (old Krakens)
  --vendor <id>           Filter devices by vendor id
  --product <id>          Filter devices by product id
  --release <number>      Filter devices by release number
  --serial <number>       Filter devices by serial number
  --bus <bus>             Filter devices by bus
  --address <address>     Filter devices by address in bus
  --usb-port <port>       Filter devices by USB port in bus
  --pick <number>         Pick among many results for a given filter
  -v, --verbose           Output additional information
  -g, --debug             Show debug information on stderr
  --version               Display the version number
  --help                  Show this message

Examples:
  liquiddump
  liquiddump --product 0xb200
  liquiddump --interval 0.5
  liquiddump > file.jsonl
  liquiddump | jq -c .

Copyright Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import json
import logging
import sys
import time

import liquidctl.cli as _borrow
import usb
from docopt import docopt
from liquidctl.driver import *

LOGGER = logging.getLogger(__name__)

if __name__ == "__main__":
    args = docopt(__doc__, version="0.1.1")
    frwd = _borrow._make_opts(args)
    devs = list(find_liquidctl_devices(**frwd))
    update_interval = float(args["--interval"])

    if args["--debug"]:
        args["--verbose"] = True
        logging.basicConfig(level=logging.DEBUG, format="[%(levelname)s] %(name)s: %(message)s")
    elif args["--verbose"]:
        logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
    else:
        logging.basicConfig(level=logging.WARNING, format="%(message)s")
        sys.tracebacklimit = 0

    try:
        for d in devs:
            LOGGER.info("initializing %s", d.description)
            d.connect()
        status = {}
        while True:
            for d in devs:
                try:
                    status[d.description] = d.get_status()
                except usb.core.USBError as err:
                    LOGGER.warning("failed to read from the device, possibly serving stale data")
                    LOGGER.debug(err, exc_info=True)
            print(json.dumps(status), flush=True)
            time.sleep(update_interval)
    except KeyboardInterrupt:
        LOGGER.info("canceled by user")
    except:
        LOGGER.exception("unexpected error")
        sys.exit(1)
    finally:
        for d in devs:
            try:
                LOGGER.info("disconnecting from %s", d.description)
                d.disconnect()
            except:
                LOGGER.exception("unexpected error when disconnecting")



import pytest

from _testutils import MockHidapiDevice
from liquidctl.driver.asus_ryujin import AsusRyujin


@pytest.fixture
def mockRyujin():
    return AsusRyujin(_MockRyujinDevice(), "Mock ASUS Ryujin II")


class _MockRyujinDevice(MockHidapiDevice):
    def __init__(self):
        super().__init__(vendor_id=0x0B05, product_id=0x1988)
        self.requests = []
        self.response = None

    def write(self, data):
        super().write(data)
        self.requests.append(data)

        assert data[0] == 0xEC
        header = data[1]

        # Sampled responses without trailing zeros
        if header == 0x82:  # Get firmware info
            self.response = "ec02004155524a312d533735302d30313034"
        elif header == 0x99:  # Get cooler status (temperature, pump speed, embedded fan speed)
            self.response = "ec19001b056405100e"
        elif header == 0x9A:  # Get pump and embedded fan duty
            self.response = "ec1a0000223c"
        elif header == 0xA0:  # Get AIO fan controller fan speeds
            self.response = "ec200000000c03ee02"
        elif header == 0xA1:  # Get AIO fan controller duty
            self.response = "ec2100005b"
        elif header == 0x1A:  # Set pump and embedded fan duty
            self.response = "ec1a"
        elif header == 0x21:  # Set AIO fan controller duty
            self.response = "ec21"
        else:
            self.response = None

    def read(self, length, **kwargs):
        pre = super().read(length, **kwargs)
        if pre:
            return pre

        buf = bytearray(65)
        buf[0] = 0xEC

        if self.response:
            response = bytes.fromhex(self.response)
            buf[: len(response)] = response

        return buf[:length]


def test_initialize(mockRyujin):
    with mockRyujin.connect():
        (firmware_status,) = mockRyujin.initialize()

        assert firmware_status[1] == "AURJ1-S750-0104"


def test_status(mockRyujin):
    with mockRyujin.connect():
        actual = mockRyujin.get_status()

        expected = [
            ("Liquid temperature", pytest.approx(27.5), "°C"),
            ("Pump speed", 1380, "rpm"),
            ("Pump fan speed", 3600, "rpm"),
            ("Pump duty", 34, "%"),
            ("Pump fan duty", 60, "%"),
            ("External fan duty", 36, "%"),
            ("External fan 1 speed", 780, "rpm"),
            ("External fan 2 speed", 750, "rpm"),
            ("External fan 3 speed", 0, "rpm"),
            ("External fan 4 speed", 0, "rpm"),
        ]

    assert sorted(actual) == sorted(expected)


def test_set_fixed_speeds(mockRyujin):
    with mockRyujin.connect():
        mockRyujin.set_fixed_speed(channel="pump", duty=10)
        assert mockRyujin.device.requests[-1][3] == 0x0A

        mockRyujin.set_fixed_speed(channel="pump-fan", duty=20)
        assert mockRyujin.device.requests[-1][4] == 0x14

        mockRyujin.set_fixed_speed(channel="external-fans", duty=30)
        assert mockRyujin.device.requests[-1][4] == 0x4C

        mockRyujin.set_fixed_speed(channel="fans", duty=40)
        assert mockRyujin.device.requests[-2][4] == 0x28
        assert mockRyujin.device.requests[-1][4] == 0x66



#!/usr/bin/env python3

"""krakenduty proof of concept – translate Kraken X speeds to duty values

This is just a proof of concept.

Usage:
  krakenduty-poc train
  krakenduty-poc status
  krakenduty-poc --help
  krakenduty-poc --version

Copyright Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import ast
from time import sleep

from docopt import docopt
from liquidctl.driver.kraken_two import KrakenTwoDriver
from liquidctl.util import interpolate_profile as interpolate

DATAFILE = ".krakenduty-poc"
DUTY_STEP = 5
DUTY_SLEEP = 5
DUTY_SAMPLES = 5


def get_speeds(device):
    status = {k: v for k, v, u in device.get_status()}
    return (status["Fan speed"], status["Pump speed"])


def find_duty_values(training_data, fan_speed, pump_speed):
    # for now simply interpolate, but this is terrible because it ignores variance
    fan_duty = interpolate(sorted([(speed, duty) for duty, speed, _ in training_data]), fan_speed)
    pump_duty = interpolate(sorted([(speed, duty) for duty, _, speed in training_data]), pump_speed)
    # don't return values outside the allowed bounds to avoid confusion
    return (min(max(fan_duty, 25), 100), min(max(pump_duty, 50), 100))


def do_train(device):
    # read current values
    fan_speed, pump_speed = get_speeds(device)
    print("starting values: fan = {} rpm, pump = {} rpm".format(fan_speed, pump_speed))

    # train
    training_data = []
    for duty in range(0, 101, DUTY_STEP):
        # don't worry if duty is off spec, the driver will correct it
        for channel in ["fan", "pump"]:
            device.set_fixed_speed(channel, duty)
        # wait significantly to allow the speed to stabilize
        sleep(DUTY_SLEEP)
        # get a few samples because there is some natural variation;
        # though this might need some delays and, also, depend on the actually
        # observed variance
        samples = [get_speeds(device) for i in range(DUTY_SAMPLES)]
        average = [sum(i) / len(i) for i in zip(*samples)]
        print("{}% duty: fan = {:.0f} rpm, pump = {:.0f} rpm".format(duty, *average))
        training_data.append([duty] + average)
        with open(DATAFILE, "w") as f:
            f.write(str(training_data))

    # (try to) restore the current values
    fan_duty, pump_duty = find_duty_values(training_data, fan_speed, pump_speed)
    print("applying fixed values: fan = {}%, pump = {}%".format(fan_duty, pump_duty))
    device.set_fixed_speed("fan", fan_duty)
    device.set_fixed_speed("pump", pump_duty)


def do_status(device):
    # read training data
    training_data = []
    with open(DATAFILE, "r") as f:
        training_data = ast.literal_eval(f.read())

    # augment
    status = []
    for k, v, u in device.get_status():
        status.append((k, v, u))
        if k == "Fan speed":
            fan_duty, _ = find_duty_values(training_data, v, 0)
            status.append(("Fan duty", fan_duty, "%"))
        elif k == "Pump speed":
            _, pump_duty = find_duty_values(training_data, 0, v)
            status.append(("Pump duty", pump_duty, "%"))

    # report
    print("{}".format(device.description))
    for k, v, u in status:
        print("{:<20}  {:>6}  {:<3}".format(k, v, u))
    print("")


if __name__ == "__main__":
    args = docopt(__doc__, version="0.0.2")

    device = KrakenTwoDriver.find_supported_devices()[0]
    device.connect()
    try:
        if args["train"]:
            do_train(device)
        elif args["status"]:
            do_status(device)
        else:
            raise Exception("nothing to do")
    finally:
        device.disconnect()
